/*
 * 05/28/2003
 *
 * camerafree.cpp
 * Author: Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "camerafree.h"

// ##################
// Classe CCameraFree
// ##################


// ========================
// Constructeur
// Initialise :
// position = (0, 0, 0)
// cible = (0, 1, 0.5)
// "up" vecteur = (0, 0, 1)
// ========================
CCameraFree::CCameraFree(id_t id, char * name) : CCamera(id, name)
{
	m_type = CAMERA_FREE;
}


// ===========
// Destructeur
// ===========
CCameraFree::~CCameraFree() { }


// ============================================================
// Effectue une rotation de l'angle de vue par rapport a un axe
// ============================================================
void CCameraFree::RotateView(float angle, float x, float y, float z)
{
	CVector vNewView;

	// Calcul le vecteur direction (la direction dans laquelle la camera regarde)
	CVector vView = m_vView - m_vPosition;		

	// Calcul le sinus et le cosinus de l'angle
	float cosTheta = (float)cos(angle);
	float sinTheta = (float)sin(angle);

	// http://astronomy.swin.edu.au/~pbourke/geometry/rotate/
	// Calcul le nouveau vecteur representant la direction de la camera

	// Find the new x position for the new rotated point
	vNewView.x  = (cosTheta + (1 - cosTheta) * x * x)		* vView.x;
	vNewView.x += ((1 - cosTheta) * x * y - z * sinTheta)	* vView.y;
	vNewView.x += ((1 - cosTheta) * x * z + y * sinTheta)	* vView.z;

	// Find the new y position for the new rotated point
	vNewView.y  = ((1 - cosTheta) * x * y + z * sinTheta)	* vView.x;
	vNewView.y += (cosTheta + (1 - cosTheta) * y * y)		* vView.y;
	vNewView.y += ((1 - cosTheta) * y * z - x * sinTheta)	* vView.z;

	// Find the new z position for the new rotated point
	vNewView.z  = ((1 - cosTheta) * x * z - y * sinTheta)	* vView.x;
	vNewView.z += ((1 - cosTheta) * y * z + x * sinTheta)	* vView.y;
	vNewView.z += (cosTheta + (1 - cosTheta) * z * z)		* vView.z;

	// On additionne le nouveau vecteur de direction a la position
	// de la camera pour avoir la nouvelle cible de la camera
	m_vView = m_vPosition + vNewView;
}


// ===========================================================================================
// Deplace la camera ainsi que la cible vers la gauche ou la droite suivant la valeur de speed
// ===========================================================================================
void CCameraFree::StrafeCamera(float speed)
{
	// On calcule le "strafe" vecteur qui est perpendiculaire au vecteur direction et "up"
	CVector vStrafe = (m_vView - m_vPosition) ^ m_vUpVector;

	// Normalisation de "strafe"
	vStrafe = CVector::Normalize(vStrafe);

	// On ajoute le "strafe" vecteur  la position de la camera
	m_vPosition.x += vStrafe.x * speed;
	m_vPosition.z += vStrafe.z * speed;

	// On ajoute le "strafe" vecteur  la position de la cible de la camera
	m_vView.x += vStrafe.x * speed;
	m_vView.z += vStrafe.z * speed;
}


// =========================================================================================
// Deplace la camera ainsi que la cible vers l'avant ou l'arriere suivant la valeur de speed
// =========================================================================================
void CCameraFree::MoveCamera(float speed)
{
	// Calcul le vecteur direction (la direction dans laquelle la camera regarde)
	CVector vVector = m_vView - m_vPosition;

	// Normalisation du vecteur de direction
	vVector = CVector::Normalize(vVector);

	// On ajoute le vecteur de direction  la position de la camera
	m_vPosition.x += vVector.x * speed;
	m_vPosition.y += vVector.y * speed;
	m_vPosition.z += vVector.z * speed;

	// On ajoute le vecteur de direction  la position de la cible de la camera
	m_vView.x += vVector.x * speed;
	m_vView.y += vVector.y * speed;
	m_vView.z += vVector.z * speed;
}


// ==============================================================================
// Change l'angle de vue de la camera en fonction des deplacements de la souris
// x,y correspondent a la position courante de la souris
// lx,ly correspondent a la derniere position de la souris (avant le deplacement)
// ==============================================================================
void CCameraFree::SetViewByMouse(int x, int y, int lx, int ly, float speed)
{
	float angleY = 0.0f;
	float angleZ = 0.0f;
	static float currentRotX = 0.0f;
	
	// si la souris n'a pas bouge on ne fait rien
	if( (x == lx) && (y == ly) ) return;

	// On determine les angles de rotation suivant les deplacements
	// de la souris et la vitesse "speed"
	angleY = (float)( (lx - x) ) / speed;		
	angleZ = (float)( (ly - y) ) / speed;
	
	// Ici on garde la rotation suivant X
	currentRotX -= angleZ;  
	
	// On teste la rotation X pour eviter de faire des rotations de plus de 180
	if(currentRotX > 1.4f)
		currentRotX = 1.4f;
	else if(currentRotX < -1.4f)
		currentRotX = -1.4f;

	else
	{
		// La rotation X (haut et bas) correspond a une rotation autour
		// de l'axe perpendiculaire au vecteur direction et au vecteur "up"
		CVector vAxis = (m_vView - m_vPosition) ^ m_vUpVector;
		vAxis = CVector::Normalize(vAxis);

		// Rotation autour de l'axe perpendiculaire (haut, bas)
		// et autour de l'axe Y (gauche, droite)
		RotateView(angleZ, vAxis.x, vAxis.y, vAxis.z);
		RotateView(angleY, 0, 1, 0);
	}
}


// ===================================================================
// Deplace la camera ainsi que la cible en fonction de la touche "key"
// ===================================================================
void CCameraFree::SetViewByKeyboard(int key, float speed)
{
	switch(key) 
	{
		case GLUT_KEY_UP:
			MoveCamera(speed);
			break;
		case GLUT_KEY_DOWN:
			MoveCamera(-speed);
			break;
		case GLUT_KEY_RIGHT:
			StrafeCamera(speed);
			break;
		case GLUT_KEY_LEFT:
			StrafeCamera(-speed);
			break;
	}
}


// =================
// Dessine la camera
// =================
void CCameraFree::Draw(CFrustum * frustum)
{
	glBindTexture(GL_TEXTURE_2D, 0);
	glPushMatrix();
	glColor3f(1.0,1.0,1.0);
	glBegin(GL_LINES);
		glVertex3f(m_vPosition.x, m_vPosition.y, m_vPosition.z);
		glVertex3f(m_vView.x, m_vView.y, m_vView.z);
	glEnd();
	glTranslatef(m_vPosition.x, m_vPosition.y, m_vPosition.z);
	glutWireSphere(0.5, 10, 10);
	glPopMatrix();
}
